 /*
		complex functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>

#ifndef _COMPLEX_H_
#define _COMPLEX_H_       

typedef struct Complex {
	double real;
	double imag;
} Complex;

////////////////////////////////////////////////////////////////////////////////////////////////////////////

//static const Complex CMinusOne = {-1., 0.};
static const Complex CZero = {0., 0.};
static const Complex COneHalf = {.5, 0.};
static const Complex COne = {1., 0.};
static const Complex CTwo = {2., 0.};
static const Complex CE = {2.7182818284590452, 0.};
static const Complex CHalfPi = {1.5707963267948966, 0.};
static const Complex CPi = {3.1415926535897932, 0.};
static const Complex CTwoPi = {6.2831853071795864, 0.};

////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* add two complex numbers */
Complex CAdd(Complex a, Complex b);

/* add two complex numbers */
Complex CSub(Complex a, Complex b);

/* multiply two complex numbers */
Complex CMul(Complex a, Complex b);

/* divide two complex numbers */
Complex CDiv(Complex a, Complex b);

/* negate complex number */
Complex CNeg(Complex z);

/* conjugate of complex number */
Complex CConj(Complex a);

/* magnitude of complex number */
double CAbs(Complex a);

/* phase of complex number */
double CAngle(Complex a); 

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* multiply real with complex */
Complex CMulReal(double r, Complex c);  

/* divide complex by real*/
Complex CDivReal(double r, Complex c);  

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* convert double to complex */
Complex Double2Comp(double x);

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* complex arc cosine */
Complex CACos(Complex z);

/* complex hyperbolic arc cosine */
Complex CACosH(Complex z);

/* complex arc sine */
Complex CASin(Complex z);

/* complex hyperbolic arc sine */
Complex CASinH(Complex z);

/* complex arc tangent */
Complex CATan(Complex z);

/* complex hyperbolic arc tangent */
Complex CATanH(Complex z);

/* complex cosine */
Complex CCos(Complex z);

/* complex hyperbolic cosine */
Complex CCosH(Complex z);

/* complex exponential */
Complex CExp(Complex z);

/* complex natural logarithm (base e) */
Complex CLn(Complex z);

/* complex decadic logarithm (base 10) */
Complex CLog10(Complex z);

/* complex power function x^y
https://de.wikipedia.org/wiki/Komplexe_Zahl#Beliebige_komplexe_Exponenten */
Complex CPow(Complex x, Complex y);

/* complex sine */
Complex CSin(Complex z);

/* complex hyperbolic sine */
Complex CSinH(Complex z);

/* complex square root */
Complex CSqrt(Complex z);

/* complex tangent */
Complex CTan(Complex z);

/* complex hyperbolic tangent */
Complex CTanH(Complex z);

#endif

