/*
		Gamma function
		
    Algorithm Copyright (C) 2015  Lanczos approximation from Wikipedia
    Port from python to C Copyright (C) 2015  Ralph Schenn
	
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Stack.h"

#define LEN 8

/* real gamma function 
would also work for complex case when using complex instead of real functions
argument of gamma function
http://dlmf.nist.gov/5.11#SS1.p2 */
double Gamma(double z)
{  
	Access_AMS_Global_Variables; 

	const double p[] = {
		676.5203681218851,
		-1259.1392167224028,
		771.32342877765313,
		-176.61502916214059,
		12.507343278686905,
		-0.13857109526572012,
		9.9843695780195716e-6,
		1.5056327351493116e-7
	};

	double t, x;
	unsigned char i; 
	    
   	// Reflection formula
	//if (__real__ z < 0.5) {
	if (z < 0.5) {
		return FLOATTAB[FPI_ONEPI] / (sin(FLOATTAB[FPI_ONEPI] * z) * Gamma(1 - z));
    } else {
		z -= 1;
		x = 0.99999999999980993;
 
		for (i = 0; i < LEN; i++) {
			x += p[i] / (z + i + 1);
		}
 
		t = z + LEN - 0.5;
		return sqrt(2 * FLOATTAB[FPI_ONEPI]) * pow(t, z + 0.5) * exp(-t) * x;
	}
}
/*Complex Gamma(Complex z)
{  
	unsigned char i; 

	const Complex p[] = {
		{676.5203681218851, 0.},
		{-1259.1392167224028, 0.},
		{771.32342877765313, 0.},
		{-176.61502916214059, 0.},
		{12.507343278686905, 0.},
		{-0.13857109526572012, 0.},
		{9.9843695780195716e-6, 0.},
		{1.5056327351493116e-7, 0.}
	};

	Complex t, x;
		
   	// Reflection formula
	if (z.real < 0.5) {
		return CDiv (CPi, (CMul (CSin (CMul (CPi, z)), Gamma (CSub (COne, z)))));
    } else {
		z = CSub (z, COne);
		
		x.real = 0.99999999999980993;
		x.imag = 0;
 
		for (i = 0; i < LEN; i++) {
			x = CAdd (x, CDiv (p[i], (CAdd (CAdd (z, Double2Comp (i)), COne))));
		}
 
		t = CSub (CAdd (z, Double2Comp (LEN)), COneHalf);
		return CMul (CMul (CMul (CSqrt (CMul (CTwo, CPi)), CPow (t, CAdd (z, COneHalf))), CExp (CNeg (t))), x);
	}
}*/

/* push gamma function */
void push_gamma (EStackIndex i)
{
	// push value of function as result
	push_Float (Gamma (estack_number_to_Float_inf (i)));
}
