/*
		multi expressions
		
    Copyright (C) 2013  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "MultiExpr.h"
#include "ROM.h"
#include "Search.h"

#define MAX_NAME_LENGTH 9
#define MAX_PATH_LENGTH 2 * MAX_NAME_LENGTH + 1
#define PATH_SEPAR '\\'
 
/* extract reference from source code 
returns null-terminated string with the format folder\\variable */
HANDLE GetPath(const char *expr, 
	unsigned short length, 
	unsigned short backSlash) // ok
{    
	// get begin of path
	unsigned short begin = PathBegin(expr, backSlash);
    unsigned short end;
    unsigned short pathLength;
    HANDLE path;
    char *pathPtr;
    
	// begin could not be found
	if (begin == 0) {
		return H_NULL;
	}
        
	// get end of path
	end = PathEnd(expr, length, backSlash);
  
	// begin could not be found
	if (end == 0) {
		return H_NULL;
	}

	// length of path
	pathLength = end - begin + 1;
  
	// the shortest possible path is x\y
	if (pathLength > 2) {
		// allocate space for reference string (path)
		path = HeapAllocThrow (pathLength);
	        
		// get pointer to reference string
		pathPtr = HeapDeref (path);
	
		// copy substring from function/program to allocated block
		strncpy (pathPtr, expr + begin, pathLength); 
	  
		// set string terminator
		pathPtr[pathLength] = '\0';
	        
		return path;  	
	}
  
	return H_NULL;
}
 
/* returns array of rferences (strings with calls to user defined 
functions/programs contained in variable 
the format of the strings is folder\\variable */
/*unsigned short GetReferences(HANDLE *refs, 
	SYM_ENTRY *symptr) // ok
{
	// characters that have already been searched
	unsigned short searched = 0;
        
	// number of found references
	unsigned short numRefs = 0;
  
	const char *separator;
    HANDLE ref;
    HANDLE *ptr;
    unsigned short idx;
    MULTI_EXPR *expr;
    
	// variable is a function or program
 	if (IsFuncOrProg(symptr)) {
		// get pointer to structure describing variable
		expr = HeapDeref (symptr->handle);
	  
		if (expr != NULL) {
			// begin of source code
			separator = expr->Expr;
		        	
			// more source code to search
			// (this is text, not tagged representation of function/program)
			while (searched < expr->Size) {
				// search the next path separator
				separator = memchr (separator, PATH_SEPAR, expr->Size - searched);
			                
				// found next occurence of path separator 
				if (separator != NULL) {
					// get handle to reference of actual separator
					ref = GetPath(expr->Expr, expr->Size, (ESQ *)separator - expr->Expr);
			 	  
					// begin and end of rference was found
					if (ref != H_NULL &&
			 	  	// this path has not been aready added to the array
						ArrayContStr(HeapDeref (*refs), numRefs, HeapDeref (ref)) == -1) {
			    		numRefs++;
		
						// allocate array for handles to references (strings with paths)
						*refs = HeapRealloc (*refs, numRefs * sizeof (HANDLE));
		      
						if (*refs == H_NULL) {
							ER_throwVar (FIRST_MEMORY_ERROR);
						}
						
						// get pointer to array with handles to references (strings with paths)
						ptr = HeapDeref (*refs);
		    	
						// index of next reference
						idx = numRefs - 1;
		    	
						ptr[idx] = H_NULL;
		    		
						// get pointer to reference (path string)
						ptr[idx] = ref;	 	
					}
			      
					// start next search after previous search
					// (otherwise we will find the same path separator again)
					// we could increment by length of var name here
					separator++;
					// no more occurences of path separator
				} else {
					break;
				}
			}
		}   
	}
  
  return numRefs;
}*/

/* check if variable is a file (not a folder) */
Boolean IsFile(SYM_ENTRY *symptr) // ok
{
	return !(symptr->Flags & SF_FOLDER);
}

/* check if variable is a function or program */
/*Boolean IsFuncOrProg(SYM_ENTRY *symptr) // ok
{
	return (((MULTI_EXPR *)(HeapDeref (symptr->hVal)))->Expr)[0] == '('; // ???
}*/

/* check if expr contains str
str must be null-terminated */
Boolean MultiExprContStr(MULTI_EXPR *expr, 
	const char *str) // ok
{
	return BlockContBlock((const char *)expr->Expr, 
		str, 
		expr->Size, 
		strlen (str)) != NULL;
}

/* find begin of reference path */
unsigned short PathBegin(const char *expr, 
	unsigned short backSlash) // ok
{
  #define BEGIN_LENGTH 23

	char beginChars[] = { 
		' ',
		'\r',   // carriage return
		'\0',   // colon
		'(',
		'[',    // matrix
		'{',    // list
		'<',
		'\x3F', // <=
		'>',
		'\x9E', // >=
		'=',
		'\x9D', // !=
		',',    // parameter divider
		'\xAD', // negative sign
		'+',
		'-',
		'*',
		'/',
		'^',
		'\x9F', // angle
		'&',    // string concat
		'"',    // function call (as string)
		'\xA9'  // comment
	};
  
  const char *ptr;
        
	// search begin of path
	// stop after walking MAX_NAME_LENGTH to the begin
	for (ptr = expr + backSlash - 1; ptr > expr; ptr--) {
		// begin of comment (c)
		if ((unsigned char)(*ptr) == 0xA9) {
			return 0;
		// begin of reference
		} else if (memchr (beginChars, *ptr, BEGIN_LENGTH) != NULL) {
			return ptr - expr + 1;
		}
	} 
  
	// begin was not found
	return 0;
}
 
/* find end of reference path */
unsigned short PathEnd(const char *expr, 
	unsigned short length, 
	unsigned short backSlash) // ok
{
	#define NOT_END_LENGTH 18

	char notEndChars[] = { 
		' ',
		'\r',   // carriage return
		'\0',   // colon
		'[',    // matrix
		'<',
		'\x3F', // <=
		'>',
		'\x9E', // >=
		'=',
		'\x9D', // !=
		',',    // parameter divider
		'+',
		'-',
		'*',
		'/',
		'^',
		'&',    // string concat
		//'"',    // function call (as string)
		'\xA9'  // comment
	};
  
	const char *ptr, *end = expr + length;
  
	// search end of path
	for (ptr = expr + backSlash + 1; ptr < end; ptr++) {
		// end was found
		if (*ptr == '(') {
			return ptr - expr - 1;
		// other character was found
		} else if (memchr (notEndChars, *ptr, NOT_END_LENGTH) != NULL) {
			return 0;
		}
} 
  
  // end of reference was not found
  return 0;
}
 
/* variable contains string */
/*Boolean SymEntryContStr(SYM_ENTRY *symptr, 
	const char *str) // 
{
	// get pointer to structure describing variable
  MULTI_EXPR *expr = HeapDeref (symptr->handle);
  
  // HeapDeref was successfull
  if (expr != NULL)
  {
	  // string was found
	  return MultiExprContStr(expr, str);
	}
	
	return FALSE;
}*/

/* check if called function/program exists */
/*Boolean VarExists(const char *path) // ok
{
  // search the variable with path
  //HSYM var = SymFind (SYMSTR (path));
  HSYM var = SymFind (SYMSTR (path));
        
  // variable was not found
  return (var.folder != 0) ? TRUE : FALSE;
}*/
