/*
		string functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "String.h"

/* check if s1 contains s2 
case insensitive version 
function is non-destructive 
caution: if you can avoid dynamic memory alloc and making copies of strings 
you will get much better performance (using strtolower and strstr) */
char * stristr (const char * s1, const char * s2)
{
	// length of strings
	unsigned short l1, l2;
	
	// handles/pointers for lower case copy of strings
	volatile HANDLE h1, h2; 
	char *p1 = NULL, *p2, *p = NULL;
	       
	l1 = strlen (s1) + 1;
	l2 = strlen (s2) + 1;
	
	TRY
		// allocate spaces for string copies
		h1 = HeapAllocThrow (l1);
		h2 = HeapAllocThrow (l2);
		
		p1 = (char *)HeapDeref (h1);
		p2 = (char *)HeapDeref (h2);
		
		// copy the strings
		memcpy (p1, s1, l1);
		memcpy (p2, s2, l2);
		
		// convert to lower case
		strtolower (p1, p1);
		strtolower (p2, p2);
		
		// test if p1 conatins p2
		p = strstr (p1, p2);
	FINALLY
	    HeapFreeIndir ((HANDLE *)&h1);
	    HeapFreeIndir ((HANDLE *)&h2);
	ENDFINAL

	// if p != NULL it has to be >= p1
	// the return pointer must be relative to s1, not p1
	// because the caller knows only s1 not its copy p1
	return p != NULL ? (char *)s1 + (p - p1) : NULL;
}

/* convert string to lower case
dst == src: destructive conversion 
dst != src : nondestructive conversion 
memory allocation must be handled by the caller */
void strtolower(char *dst, char *src) {
    Access_AMS_Global_Variables;  

	char *d = dst, *s = src;
	 
	// loop over characters of source string
	for ( ; *s; s++) {
		// copy lower case character to destination
		// tolower knows which characters are alphabetical
		// so we must must not check using is_alpha
		*d++ = tolower(*s);
	}       
	
	// add termination
	*d = '\0';
}
                                         
