/*
		convert tag decimal values to strings with tag names
		
    Copyright (C) 2013  Ralph Schenn with support from Kevin Kofler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Stack.h"
#include "String.h"
#include "Tags.h"
 
#define BUF_LEN 25

// tags
static char *tags[] = {
	"VAR_NAME",     // 0
	"PRIV_VAR",
	"VAR_R",
	"VAR_S",
	"VAR_T",
	"VAR_U",
	"VAR_V",
	"VAR_W",
	"VAR_X",
	"VAR_Y",
	"VAR_Z",
	"VAR_A",
	"VAR_B",
	"VAR_C",
	"VAR_D",
	"VAR_E",
	"VAR_F",
	"VAR_G",
	"VAR_H",
	"VAR_I",
	"VAR_J",
	"VAR_K",
	"VAR_L",
	"VAR_M",
	"VAR_N",
	"VAR_O",
	"VAR_P",
	"VAR_Q",        
	"EXT_SYS",
	"ARB_REAL",
	"ARB_INT",
	"POSINT",
	"NEGINT",
	"POSFRAC",
	"NEGFRAC",
	"FLOAT",
	"PI",
	"EXP",
	"IM",
	"NEGINF",
	"INFINITY",
	"PN_INF",
	"UNDEF",
	"FALSE",
	"TRUE",
	"STR",
	"NOTHING",
	"ACOSH",
	"ASINH",
	"ATANH",
	"ASECH",
	"ACSCH",
	"ACOTH",
	"COSH",
	"SINH",
	"TANH",
	"SECH",
	"CSCH",
	"COTH",
	"ACOS",
	"ASIN",
	"ATAN",
	"ASEC",
	"ACSC",
	"ACOT",
	"RACOS",
	"RASIN",
	"RATAN",
	"COS",
	"SIN",
	"TAN",
	"SEC",
	"CSC",
	"COT",
	"ITAN",
	"ABS",
	"ANGLE",
	"CEILING",
	"FLOOR",
	"INT",
	"SIGN",
	"SQRT",
	"EXPF",
	"LN",
	"LOG",
	"FPART",
	"IPART",
	"CONJ",
	"IMAG",
	"REAL",
	"APPROX",
	"TEXPAND",
	"TCOLLECT",
	"GETDENOM",
	"GETNUM",
	"ERROR",
	"CUMSUM",
	"DET",
	"COLNORM",
	"ROWNORM",
	"NORM",
	"MEAN",
	"MEDIAN",
	"PRODUCT",
	"STDDEV",
	"SUM",
	"VARIANCE",
	"UNITV",
	"DIM",
	"MAT2LIST",
	"NEWLIST",
	"RREF",
	"REF",
	"IDENTITY",
	"DIAG",
	"COLDIM",
	"ROWDIM",
	"TRANSP",
	"FACT",
	"PERCENT",
	"RADIANS",
	"NOT",
	"MINUS",
	"VEC_POL",
	"VEC_CYL",
	"VEC_SPH",
	"START",
	"ISTORE",
	"STORE",
	"WITH",
	"XOR",
	"OR",
	"AND",
	"LT",
	"LE",
	"EQ",
	"GE",
 	"GT",
	"NE",
	"ADD",
	"ADDELT",
	"SUB",
	"SUBELT",
	"MUL",
	"MULELT",
	"DIV",
	"DIVELT",
	"POW",
	"POWELT",
	"SINCOS",
	"SOLVE",
	"CSOLVE",
	"NSOLVE",
	"ZEROS",
	"CZEROS",
	"FMIN",
	"FMAX",
	"COMPLEX",
	"POLYEVAL",
	"RANDPOLY",
	"CROSSP",
	"DOTP",
	"GCD",
	"LCM",
	"MOD",
	"INTDIV",
	"REMAIN",
	"NCR",
	"NPR",
	"P2RX",
	"P2RY",
	"P2PTHETA",
	"P2PR",
	"AUGMENT",
	"NEWMAT",
	"RANDMAT",
	"SIMULT",
	"PART",
	"EXP2LIST",
	"RANDNORM",
	"MROW",
	"ROWADD",
	"ROWSWAP",
	"ARCLEN",
	"NINT",
	"PI_PROD",
	"SIG_SUM",
	"MROWADD",
	"ANS",
	"ENTRY",
	"EXACT",
	"LOGB",
	"COMDENOM",
	"EXPAND",
	"FACTOR",
	"CFACTOR",
	"INTEG",
	"DIFFEREN",
	"AVGRC",
	"NDERIV",
	"TAYLOR",
	"LIMIT",
	"PROPFRAC",
	"WHEN",
	"ROUND",
	"DMS",
	"LEFT",
	"RIGHT",
	"MID",
	"SHIFT",
	"SEQ",
	"LIST2MAT",
	"SUBMAT",
	"SUBSCR",
	"RAND",
	"MIN",
	"MAX",
	"LIST",
	"USERFUNC",
	"MATRIX",
	"FUNC",
	"DATA",
	"GDB",
	"PIC",
	"TEXT",
	"FIG",
	"MAC",
	"EXT",
	"EXT_INST",
	"END",
	"COMMENT",
	"NEXTEXPR",
	"NEWLINE",
	"ENDSTACK",
	"PN1",
	"PN2",
	"ERR_MSG",
	"EIGVC",
	"EIGVL",
	"DASH",
	"LOCALVAR",
	"DESOLVE",
	"FDASH",
	"ASM",
	"ISPRIME",
	"OTH",
	"ROTATE" // 249
};
 
char *Tag2Str(unsigned char tag)
{
	return tags[tag]; 
}
 
//#define MAX_TAG 249
 
/* convert string to tag */
/*enum Tags Str2Tag(const char *input)
{       
	unsigned char i;
        
	for (i = 0; i < MAX_TAG; i++) {
		// the string was found
		if (strcmp(tags[i], input) == 0) {
			return --i;
		}
	}
        
	// the string was not found
	ER_throw (ER_ARGUMENT);
}*/
 
// extended tags
static char *extTags[]  = {
	"INDIR", // extTags[0] = 1
	"GETKEY",
	"GETFOLD",
	"SWITCH",
	"UNITCONV",
	"ORD",
	"EXPR",
	"CHAR",
	"STRING",
	"GETTYPE",
	"GETMODE",
	"SETFOLD",
	"PTTEST",
	"PXLTEST",
	"SETGRAPH",
	"SETTABLE",
	"SETMODE",
	"FORMAT",
	"INSTRING",
	"APPEND",
	"DD",
	"EXPR2DMS",
	"VEC2RECT",
	"VEC2POL",
	"VEC2CYL",
	"VEC2SPH",
	"PARENSTA",
	"PARENEND"
	"MAT_STA",
	"MAT_END",
	"LIST_STA",
	"LIST_END",
	"COMMA",
	"SEMICOL",
	"COMP_ANG",
	"SING_QUO",
	"QUOTE",
	"POLCPLX",
	"TMPCNV",
	"DEL_TCNV",
	"GETUNITS",
	"SETUNITS",
	"BIN",
	"HEX",
	"INT2BIN",
	"INT2DEC",
	"INT2HEX",
	"DET_TOL",
	"REF_TOL",
	"RREF_TOL",
	"SIMULTOL",
	"GETCONFG",
	"V_AUGMEN",
	"MEAN2ARG",
	"PROD2ARG",
	"SDEV2ARG",
	"SUM_2ARG",
	"VAR_2ARG",
	"DELTALST", // extTag[58] = 59
	
	"ISCLKON", // extTag[59] = 70
	"GETDATE",
	"GETTIME",
	"GETTMZN",
	"SETDATE",
	"SETTIME",
	"SETTMZN",
	"DAYOFWK",
	"STARTTMR",
	"CHECKTMR",
	"TIMECNV",
	"GETDTFMT",
	"GETTMFMT",
	"GETDTSTR",
	"GETTMSTR",
	"SETDTFMT",
	"SETTMFMT",
	"ROOT", 
	"EXPRIO", 
	"IMPDIF", 
	"STDEVPOP", 
	"ISVAR", 
	"ISLOCKED", 
	"ISARCHIV", 
	"GRAD", 
	"TO_GRAD", 
	"TO_RAD",
	"TO_LN", 
	"TOLOGBAS" // extTags[87] = 98
};
 
#define LOWER_MIN_VAL 1
#define LOWER_MAX_VAL 59
#define UPPER_MIN_VAL 70
#define UPPER_MAX_VAL 98
 
#define LOWER_MIN_IDX 0
#define LOWER_MAX_IDX 58
#define UPPER_MIN_IDX 59
#define UPPER_MAX_IDX 87
 
/* convert extended tag to string */
static char *ExtTag2Str(unsigned char tag) // ???
{
	// tag is falling in lower interval
	if (LOWER_MIN_VAL <= tag && tag <= LOWER_MAX_VAL) {
		return extTags[tag - 1]; 
	// tag is falling in upper interval
	} else if (UPPER_MIN_VAL <= tag && tag <= UPPER_MAX_VAL) {
		// mind the gap between the intervals
		return extTags[tag - 12]; 
	} else {
		ER_throwVar (ER_DOMAIN);
	}
}
 
/* convert string to extended tag */
/*enum ExtTags Str2ExtTag(const char *input)
{
	short i;
                
	// upper values are not used for extended tags
	for (i = 0; i < UPPER_MAX_IDX; i++) {
		// the string was found
		if (strcmp(extTags[i], input) == 0) {
			// string is falling in lower interval
			if (LOWER_MIN_IDX <= i && i <= LOWER_MAX_IDX) {
				// enum is not 0-based
				return i + 1;
			// string is falling in upper interval
			} else if (UPPER_MIN_IDX <= i && i <= UPPER_MAX_IDX) {
				// observe the gap between the intervals
				return i + 12;  
			}
		}
	}
        
	// the string was not found
	ER_throw (ER_ARGUMENT);
}*/
 
/*static char *cmdTags[] = {
	"CLRDRAW", // 1
	"CLRGRAPH", 
	"CLRHOME", 
	"CLRIO", 
	"CLRTABLE", 
	"CUSTOM", 
	"CYCLE", 
	"DIALOG", 
	"DISPG", 
	"DISPTBL", 
	"ELSE", 
	"ENDCUSTM", 
	"ENDDLOG", 
	"ENDFOR", 
	"ENDFUNC", 
	"ENDIF", 
	"ENDLOOP", 
	"ENDPRGM", 
	"ENDTBAR", 
	"ENDTRY", 
	"ENDWHILE", 
	"EXIT", 
	"FUNC", 
	"LOOP", 
	"PRGM", 
	"SHOWSTAT", 
	"STOP", 
	"THEN", 
	"TOOLBAR", 
	"TRACE", 
	"TRY", 
	"ZOOMBOX", 
	"ZOOMDATA", 
	"ZOOMDEC", 
	"ZOOMFIT", 
	"ZOOMIN", 
	"ZOOMINT", 
	"ZOOMOUT", 
	"ZOOMPREV", 
	"ZOOMRCL", 
	"ZOOMSQR", 
	"ZOOMSTD", 
	"ZOOMSTO", 
	"ZOOMTRIG", 
	"DRAWFUNC", 
	"DRAWINV", 
	"GOTO", 
	"LBL", 
	"GET", 
	"SEND", 
	"GETCALC", 
	"SENDCALC", 
	"NEWFOLD", 
	"PRINTOBJ", 
	"RCLGDB", 
	"STOGDB", 
	"ELSEIF", 
	"IF", 
	"IFTHEN", 
	"RANDSEED", 
	"WHILE", 
	"LINETAN", 
	"COPYVAR", 
	"RENAME", 
	"STYLE", 
	"FILL", 
	"REQUEST", 
	"POPUP", 
	"PTCHG", 
	"PTOFF", 
	"PTON", 
	"PXLCHG", 
	"PXLOFF", 
	"PXLON", 
	"MOVEVAR", 
	"DROPDOWN", 
	"OUTPUT", 
	"PTTEXT", 
	"PXLTEXT", 
	"DRAWSLP", 
	"PAUSE", 
	"RETURN", 
	"INPUT", 
	"PLOTSOFF", 
	"PLOTSON", 
	"TITLE", 
	"ITEM", 
	"INPUTSTR", 
	"LINEHORZ", 
	"LINEVERT", 
	"PXLHORZ", 
	"PXLVERT", 
	"ANDPIC", 
	"RCLPIC", 
	"RPLCPIC", 
	"XORPIC", 
	"DRAWPOL", 
	"TEXT", 
	"ONEVAR", 
	"STOPIC", 
	"GRAPH", 
	"TABLE", 
	"NEWPIC", 
	"DRAWPARM", 
	"CYCLEPIC", 
	"CUBICREG", 
	"EXPREG", 
	"LINREG", 
	"LNREG", 
	"MEDMED", 
	"POWERREG", 
	"QUADREG", 
	"QUARTREG", 
	"TWOVAR", 
	"SHADE", 
	"FOR", 
	"CIRCLE", 
	"PXLCRCL", 
	"NEWPLOT", 
	"LINE", 
	"PXLLINE", 
	"DISP", 
	"FNOFF", 
	"FNON", 
	"LOCAL", 
	"DELFOLD", 
	"DELVAR", 
	"LOCK", 
	"PROMPT", 
	"SORTA", 
	"SORTD", 
	"UNLOCK", 
	"NEWDATA", 
	"DEFINE", 
	"ELSE_TRY", 
	"CLRERR", 
	"PASSERR", 
	"DISPHOME", 
	"EXEC", 
	"ARCHIVE", 
	"UNARCHIV", 
	"LU", 
	"QR", 
	"BLDDATA", 
	"DRWCTOUR", 
	"NEWPROB", 
	"SINREG", 
	"LOGISTIC", 
	"CUSTMON", 
	"CUSTMOFF", 
	"SENDCHAT", 
	"REQUEST_THREEARG", 
	"CLOCKON", 
	"CLOCKOFF" // 155
};*/ 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* search for primary or secondary tag 
the searched string 
lowest tag
highest tag 
type of tag list to search */
static void CatSearch(char *str,
	unsigned char min, 
	unsigned char max, 
	Quantum secCmd)
{
   	Quantum p, sc;
	unsigned char i; 
	
	// mixed case catalog name
	char mixedCat[SYM_LEN + 2];

	// mixed case catalog name
	char lowerCat[SYM_LEN + 2];

	// loop over all tags of type
	for (i = max; i >= min; i--) { 
		// set the primary tag
		switch (secCmd) {
			case SECONDARY_TAG:
				p = SECONDARY_TAG;
				break;
			case COMMAND_TAG:
				p = COMMAND_TAG;
				break; 
			// primary tag
			default:
				p = i;
				break;	
		}		

		// set the secondary tag
		sc = secCmd == SECONDARY_TAG || secCmd == COMMAND_TAG ? i : 0;
		
		if (secCmd != SYSVAR_TAG) {
			strcpy (mixedCat, (char *)get_key_ptr (p, sc));
		} else {
			GetTagStrW(i, mixedCat);
			sc = SYSVAR_TAG;
		}
		
		// convert catalog name to lower case
		strtolower (lowerCat, mixedCat);
		
		// if name of string in catalog contains the searched substring (ignore case)
		if (strstr((const char *)lowerCat, str)) {
			// beginning of row
			push_quantum (END_TAG);

			// secondary tag
			push_ushort_to_integer (sc);
						
			// primary tag
			push_ushort_to_integer (p);

			// name of tag
			push_zstr ((char *)mixedCat);     
			
			// end of row
			push_quantum (LIST_TAG);
		}
	}
}

/* recursive catalog search for function containing string */
void reccatse(void)
{
    Access_AMS_Global_Variables;
    
   	// the string passed as argument
   	char str[SYM_LEN + 1];
	    
	CheckNumArgs (1);
	
	// get pointer to folder name from estack
	// copy the string to a variable independent from estack
	// (to avoid overwriting while pushing strings to the list)		
	GetStrnArgCopy (top_estack, str, SYM_LEN + 1);
	
	// convert searched string to lower case
	strtolower (str, str);
			
	// beginning of mat
	push_quantum (END_TAG);

	// search for primary tag
	// the third arg is arbitrary (but should not be SECONDARY_TAG/COMMAND_TAG)
	CatSearch (str, PI_TAG, ROTATE_TAG, VAR_TAG);   
	
	// search for secondary tag
	CatSearch (str, INDIRECTION_TAG, SETTMFMT_TAG, SECONDARY_TAG);	
	
	// search for command tag
	CatSearch (str, CLRDRAW_TAG, LOCAL_PRGM_TAG, COMMAND_TAG);	

	// search for system variable
	CatSearch (str, SV_XBAR, SV_MAX_TAG, SYSVAR_TAG);	

	// end of mat
	push_quantum (LIST_TAG);  
	
	// at least one hit was found 
	// (otherwise we would have an empty list on top of estack)
	if (is_matrix (top_estack)) {
		// just to improve readablity by horizontal scrolling
		push_transpose_aux (top_estack, FALSE);         
	}
}

/* get tag name of system variables */
void GetTagStrW(Quantum tag, char *name)
{
    char tags[2];

	tags[0] = tag;
	tags[1] = SYSVAR_TAG; 

	GetTagStr ((EStackIndex)(tags + 1), name);
}

